import logging

logger = logging.getLogger(__name__)


class Inspector:
    def __init__(self, filename, options):
        logger.debug("Initialising Inspector writer")
        self.options = options

    def write(self, estimator):
        """Print all keys and values from estimator structure

        they include also a metadata read from binary output file
        """
        for name, value in sorted(estimator.__dict__.items()):
            # skip non-metadata fields
            if name not in {'data', 'data_raw', 'error', 'error_raw', 'counter', 'pages'}:
                line = "{:24s}: '{:s}'".format(str(name), str(value))
                print(line)
        # print some data-related statistics
        print(75 * "*")

        for page_no, page in enumerate(estimator.pages):
            print("Page {} / {}".format(page_no, len(estimator.pages)))
            for name, value in sorted(page.__dict__.items()):
                # skip non-metadata fields
                if name not in {'data', 'data_raw', 'error', 'error_raw'}:
                    line = "\t{:24s}: '{:s}'".format(str(name), str(value))
                    print(line)
            print("Data min: {:g}, max: {:g}, mean: {:g}".format(
                page.data_raw.min(), page.data_raw.max(), page.data_raw.mean()))
            print(75 * "-")

        if self.options.details:
            # print data scatter-plot if possible
            if estimator.dimension == 1 and len(self.pages) == 1:
                try:
                    from hipsterplot import plot
                    print(75 * "*")
                    print("Data scatter plot")
                    plot(estimator.data_raw)
                except ImportError:
                    logger.warning("Detailed summary requires installation of hipsterplot package")
            # print data histogram if possible
            try:
                from bashplotlib.histogram import plot_hist
                print(75 * "*")
                print("Data histogram")
                plot_hist(estimator.data_raw, bincount=70, xlab=False, showSummary=True)
            except ImportError:
                logger.warning("Detailed summary requires installation of bashplotlib package")

        return 0



import logging

import numpy as np

from pymchelper.axis import MeshAxis
from pymchelper.page import Page
from pymchelper.readers.shieldhit.binary_spec import SHBDOTagID, detector_name_from_bdotag, unit_name_from_unit_id, \
    page_tags_to_save
from pymchelper.readers.shieldhit.reader_base import SHReader, read_next_token, mesh_unit_and_name
from pymchelper.shieldhit.detector.detector_type import SHDetType
from pymchelper.shieldhit.detector.estimator_type import SHGeoType

logger = logging.getLogger(__name__)


class SHReaderBDO2019(SHReader):
    """Experimental binary format reader version >= 0.7"""

    def read_data(self, estimator, nscale=1):
        logger.debug("Reading: %s", self.filename)

        with open(self.filename, "rb") as f:
            d1 = np.dtype([('magic', 'S6'), ('end', 'S2'), ('vstr', 'S16')])

            _x = np.fromfile(f, dtype=d1, count=1)  # read the data into numpy
            logger.debug("Magic : " + _x['magic'][0].decode('ASCII'))
            logger.debug("Endiannes: " + _x['end'][0].decode('ASCII'))
            logger.debug("VerStr: " + _x['vstr'][0].decode('ASCII'))

            while f:
                token = read_next_token(f)
                if token is None:
                    break

                token_id, token_type, payload_len, raw_payload = token

                payload = [None] * payload_len
                _has_geo_units_in_ascii = False

                # decode all strings (currently there will never be more than one per token)
                if 'S' in token_type.decode('ASCII'):
                    for i, payload_entry in enumerate(raw_payload):
                        # raw payload may contain non-ASCII characters (i.e. filedate on non-English Windows OS)
                        payload[i] = payload_entry.decode('ASCII', 'replace').strip()
                else:
                    payload = raw_payload

                if payload_len == 1:
                    payload = payload[0]

                try:
                    token_name = SHBDOTagID(token_id).name
                    logger.debug("Read token {:s} (0x{:02x}) value {} type {:s} length {:d}".format(
                        token_name,
                        token_id,
                        raw_payload,
                        token_type.decode('ASCII'),
                        payload_len
                    ))
                except ValueError:
                    logger.info("Found unknown token (0x{:02x}) value {} type {:s} length {:d}, skipping".format(
                        token_id,
                        raw_payload,
                        token_type.decode('ASCII'),
                        payload_len
                    ))

                # geometry type
                if SHBDOTagID.geometry_type == token_id:
                    estimator.geotyp = SHGeoType[payload.strip().lower()]

                if SHBDOTagID.geo_n_bins == token_id:
                    estimator.x = estimator.x._replace(n=payload[0])
                    estimator.y = estimator.y._replace(n=payload[1])
                    estimator.z = estimator.z._replace(n=payload[2])

                if SHBDOTagID.geo_p_start == token_id:
                    estimator.x = estimator.x._replace(min_val=payload[0])
                    estimator.y = estimator.y._replace(min_val=payload[1])
                    estimator.z = estimator.z._replace(min_val=payload[2])

                if SHBDOTagID.geo_q_stop == token_id:
                    estimator.x = estimator.x._replace(max_val=payload[0])
                    estimator.y = estimator.y._replace(max_val=payload[1])
                    estimator.z = estimator.z._replace(max_val=payload[2])

                if SHBDOTagID.geo_unit_ids == token_id and not _has_geo_units_in_ascii:
                    estimator.x = estimator.x._replace(unit=unit_name_from_unit_id.get(payload[0], ""))
                    estimator.y = estimator.y._replace(unit=unit_name_from_unit_id.get(payload[1], ""))
                    estimator.z = estimator.z._replace(unit=unit_name_from_unit_id.get(payload[2], ""))

                # Units may also be given as pure ASCII directly from SHIELD-HIT12A new .bdo format.
                # If this is available, then use those embedded in the .bdo file, instead of pymchelper setting them.
                if SHBDOTagID.geo_units == token_id:
                    _units = payload.split(";")
                    if len(_units) == 3:
                        estimator.x = estimator.x._replace(unit=_units[0])
                        estimator.y = estimator.y._replace(unit=_units[1])
                        estimator.z = estimator.z._replace(unit=_units[2])
                        _has_geo_units_in_ascii = True

                # page(detector) type, it begins new page block
                if SHBDOTagID.detector_type == token_id:
                    # here new page is added to the estimator structure
                    estimator.add_page(Page())
                    logger.debug("Setting page.dettyp = %s (%s)", SHDetType(payload), SHDetType(payload).name)
                    estimator.pages[-1].dettyp = SHDetType(payload)

                # page(detector) data is the last thing related to page that is saved in binary file
                # at this point all other page related tags should already be processed
                if SHBDOTagID.data_block == token_id:
                    logger.debug("Setting page data = %s", np.asarray(payload))
                    estimator.pages[-1].data_raw = np.asarray(payload)

                # read tokens based on tag <-> name mapping for detector
                if token_id in detector_name_from_bdotag:
                    logger.debug("Setting detector.%s = %s", detector_name_from_bdotag[token_id], payload)
                    setattr(estimator, detector_name_from_bdotag[token_id], payload)

                # read tokens based on tag <-> name mapping for pages
                if token_id in page_tags_to_save:
                    logger.debug("Setting page.%s = %s", SHBDOTagID(token_id).name, payload)
                    setattr(estimator.pages[-1], SHBDOTagID(token_id).name, payload)

            # Loop over the file is over here
            # Check if we have differential scoring, i.e. data dimension is larger than 1:
            for page in estimator.pages:
                try:
                    page.diff_axis1 = MeshAxis(n=page.page_diff_size[0],
                                               min_val=page.page_diff_start[0],
                                               max_val=page.page_diff_stop[0],
                                               name="",
                                               unit=page.page_diff_units.split(";")[0],
                                               binning=MeshAxis.BinningType.linear)
                except AttributeError:
                    logger.debug("Lack of data for first level differential scoring")
                except IndexError:
                    logger.debug("Lack of units for first level differential scoring")

                try:
                    page.diff_axis2 = MeshAxis(n=page.page_diff_size[1],
                                               min_val=page.page_diff_start[1],
                                               max_val=page.page_diff_stop[1],
                                               name="",
                                               unit=page.page_diff_units.split(";")[1],
                                               binning=MeshAxis.BinningType.linear)
                except AttributeError:
                    logger.debug("Lack of data for second level differential scoring")
                except IndexError:
                    logger.debug("Lack of units for second level differential scoring")

            # Special treatment for MCPL detector
            for page in estimator.pages:
                if page.dettyp == SHDetType.mcpl:
                    estimator.dim = 0

            # Fix names of the axis objects for different mesh type,
            # units are directly extracted from BDO tags in 2019 format
            _, xname = mesh_unit_and_name(estimator, 0)
            _, yname = mesh_unit_and_name(estimator, 1)
            _, zname = mesh_unit_and_name(estimator, 2)
            estimator.x = estimator.x._replace(name=xname)
            estimator.y = estimator.y._replace(name=yname)
            estimator.z = estimator.z._replace(name=zname)

            # Copy the SH12A specific units into the general placeholders:
            for page in estimator.pages:
                page.unit = page.detector_unit
                # in future, a user may optionally give a more specific name in SH12A detect.dat, which then
                # may be written to the .bdo file. If the name is not set, use the official detector name instead:
                if not page.name:
                    page.name = str(page.dettyp)

            # apply basic normalization for pages with normalisation tag
            for page in estimator.pages:
                page_normalisation = getattr(page, 'page_normalized', None)
                # see pymchelper/readers/shieldhit/binary_spec.py for details on the normalisation tags
                # normalize the detectors such as dose or fluence (tagged as SH_POSTPROC_NORM or 2)
                if page_normalisation == 2:
                    page.data_raw /= np.float64(estimator.number_of_primaries)
                    page.error_raw /= np.float64(estimator.number_of_primaries)

        estimator.file_format = 'bdo2019'

        logger.debug("Done reading bdo file.")
        return True



"""
Tool for creating MC input files using user-specified tables and ranges.

2019 - Niels Bassler
"""

import argparse
import logging
import os
import shutil
import sys
from dataclasses import dataclass, field
from pathlib import Path
from shutil import copyfile
from typing import Dict, Generator, List, TypeVar, Union

logger = logging.getLogger(__name__)

# path-like type hint which supports both strings and Path objects
PathLike = TypeVar("PathLike", str, bytes, os.PathLike)


@dataclass
class Config:
    """Description needed."""

    # keys and values for contant assignments
    const_dict: Dict[str, str] = field(default_factory=dict)
    # keys and values for table assignments
    table_dict: Dict[str, List[str]] = field(default_factory=dict)
    path: Path = field(
        default_factory=Path)  # full path to this file (may be relative)

    # special const dict fields:
    files: List[str] = field(default_factory=list)
    symlinks: List[str] = field(default_factory=list)


def read_config(path: PathLike, quiet: bool = True) -> Config:
    """Description needed"""
    cfg = Config(path=Path(str(path)))

    keys: List[str] = []
    _first_line_in_table = True

    with open(path, 'r') as input_file:
        for line in input_file:
            # skip comments
            if line.startswith('#'):
                continue

            # constant assigments
            if "=" in line:
                lhs_and_rhs = line.split("=")
                cfg.const_dict[lhs_and_rhs[0].strip()] = lhs_and_rhs[1].strip()
                continue
            # check if we are starting a table
            if len(line.split()) > 0:
                if _first_line_in_table:
                    keys = line.split()
                    _first_line_in_table = False
                    # add column names
                    for key in keys:
                        cfg.table_dict[key] = []
                else:
                    # append row values
                    for key, val in zip(keys, line.split()):
                        cfg.table_dict[key].append(val)

    # replace string with comma-separated filenames with list of filenames
    if "FILES" in cfg.const_dict:
        cfg.files = [
            filename.strip() for filename in cfg.const_dict["FILES"].split(",")
        ]
    if "SYMLINKS" in cfg.const_dict:
        cfg.symlinks = [
            filename.strip()
            for filename in cfg.const_dict["SYMLINKS"].split(",")
        ]
    logger.debug(f"Config: {cfg}")
    if not quiet:
        print(f"Read config: {path!s}")
        print(f"\t Template dir relative directory: {cfg.const_dict['TDIR']}")
    return cfg


@dataclass
class McFile:
    """
    General MC single file object.
    This will be used for the template files as well as the generated output files.
    """

    path: Path = field(
        default_factory=Path)  # full path to this file (may be relative)
    symlink: bool = False  # marker if file is a symlink
    lines: List[str] = field(default_factory=list)

    @property
    def fname(self):
        """Description needed"""
        return Path(self.path).name

    def write(self):
        """Description needed"""
        with open(self.path, 'w') as f:
            f.writelines(self.lines)

    def read(self):
        """Description needed"""
        with open(self.path, 'r') as f:
            self.lines = f.readlines()

    def __post_init__(self):
        """Automatically read file contents upon object creation"""
        self.read()


@dataclass
class Template:
    """Description needed."""

    files: List[McFile] = field(default_factory=list)

    @staticmethod
    def prepare(cfg: Config) -> Generator:
        """Description needed."""
        # create a new dict, with all keys, but single unique values only:
        # this is the "current unique dictionary"
        # it represent a single line in config file
        # or a one of lines generated from loop variables
        current_dict: Dict[str, Union[str, float]] = dict(cfg.const_dict)

        # loop_keys are special keys which cover a numerical range in discrete steps.
        # here we will identify them, and for each loop_key, there will be a range setup.
        # loop_keys are identified by the "_MIN" suffix:
        loop_keys = []
        for key in cfg.table_dict.keys():
            if "_MIN" in key:
                loop_keys.append(key.strip("_MIN") + "_")

        if not cfg.table_dict:
            logger.debug(f"Serving {current_dict}")
            yield current_dict

        # loop over every items corresponding to every line in table section of the config directory
        for item in zip(*cfg.table_dict.values()):
            current_line_dict = dict(zip(cfg.table_dict.keys(), item))
            current_dict.update(current_line_dict)

            if not loop_keys:
                logger.debug(f"Serving {current_dict}")
                yield current_dict

            # now prepare the ranges for every loop_key
            for loop_key in loop_keys:
                start = float(current_line_dict[loop_key + "MIN"])
                stop = float(current_line_dict[loop_key + "MAX"])
                step = float(current_line_dict[loop_key + "STEP"])
                loop_value = start
                while loop_value <= stop:
                    # append values calculated for the loop
                    current_dict[loop_key] = loop_value
                    loop_value += step

                    # set the relative energy spread:
                    if loop_key == 'E_' and "DE_FACTOR" in current_dict:
                        _de = loop_value * float(current_dict["DE_FACTOR"])
                        # HARDCODED format for DE_
                        current_dict['DE_'] = f"{_de:.3f}"

                    logger.debug(f"Serving {current_dict}")
                    yield current_dict

    def write(self,
              working_directory: PathLike,
              cfg: Config,
              quiet: bool = True):
        """Description needed."""
        if not quiet:
            print(f"Saving generated workspace to {working_directory!s}")
        for u_dict in self.prepare(cfg=cfg):
            logger.debug(u_dict)
            _wd = u_dict["WDIR"]

            # check if any keys are in WDIR subsitutions
            for key in u_dict.keys():
                token = "${" + key + "}"
                if token in _wd:

                    _s = u_dict[key]
                    if isinstance(_s, float):
                        _s = f"{u_dict[key]:08.3f}"  # HARDCODED float format for directory string
                    _wd = _wd.replace(token, _s)
            work_dir = _wd
            if not work_dir:
                print(
                    f"Directory WDIR is not set in config file {cfg}, cannot proceed"
                )

            current_working_dir = Path(str(working_directory), work_dir)
            if current_working_dir.exists():
                if not quiet:
                    print(
                        f"Workspace {current_working_dir} directory exists, cleaning ..."
                    )
                shutil.rmtree(current_working_dir)
                current_working_dir.mkdir(parents=True)

            for tf in self.files:  # tf = template filename
                output_file_path = Path(current_working_dir, tf.fname)
                output_file_path.parent.mkdir(parents=True, exist_ok=True)
                if not tf.symlink:
                    output_file_path.touch()
                    # of = output file object
                    of = McFile(path=output_file_path)
                    for line in tf.lines:
                        for key in u_dict.keys():
                            token = "${" + key + "}"
                            while token in line:
                                _s = u_dict[key]
                                if isinstance(_s, float):
                                    _s = "{u_dict[key]:.3f}"  # HARDCODED float format for loop_keys
                                line = lreplace(line, token, _s)
                        of.lines.append(line)
                    # end loop over t.files
                    of.write()
                else:
                    try:
                        logger.debug(
                            f'Creating link {output_file_path} -> {tf.path.resolve()}'
                        )
                        output_file_path.symlink_to(target=tf.path.resolve())
                    except OSError:
                        # try copying file in case creation of symbolic links fails
                        # this may be the case i.e. for Windows users without developer mode
                        # or without administrative rights
                        copyfile(src=tf.path, dst=output_file_path)


def read_template(cfg: Config) -> Template:
    """Description needed."""
    tpl = Template()

    for filename in cfg.files:
        path = Path(cfg.const_dict['TDIR'], filename)
        if not Path(cfg.const_dict['TDIR']).is_absolute():
            path = Path(cfg.path.parent, cfg.const_dict['TDIR'], filename)
        mcfile = McFile(path=path)
        tpl.files.append(mcfile)

    for filename in cfg.symlinks:
        path = Path(cfg.const_dict['TDIR'], filename)
        if not Path(cfg.const_dict['TDIR']).is_absolute():
            path = Path(cfg.path.parent, cfg.const_dict['TDIR'], filename)
        mcfile = McFile(path=path, symlink=True)
        tpl.files.append(mcfile)

    return tpl


def lreplace(text: str, old: str, new: str) -> str:
    """
    Left adjusted replacement of string f with string r, in string s.

    This function is implemented in order to fill in data in FORTRAN77 fields,
    which are tied to certain positions on the line, i.e. subsequent values
    may not be shifted.

    Finds string f in string s and replaces it with string r, but left adjusted, retaining line length.
    If length of r is shorter than length of f, remaining chars will be space padded.
    If length of r is larger than length of f, then characters will be overwritten.
    A copy of s with the replacement is returned.
    """
    result = old
    if old in text:
        idx = text.find(old)
        replacement = new
        if len(new) < len(old):
            replacement += " " * (len(old) - len(new))
        result = text[:idx] + replacement + text[idx + len(replacement):]
    return result


def main(args=None):
    """Main function."""
    if args is None:
        args = sys.argv[1:]

    import pymchelper
    parser = argparse.ArgumentParser()
    parser.add_argument('config_path',
                        metavar="config.txt",
                        type=Path,
                        default=Path('.').absolute() / "config.txt",
                        help="config file.")
    parser.add_argument('-w',
                        '--workspace',
                        type=Path,
                        default=Path('.').absolute(),
                        help="workspace directory.")
    parser.add_argument('-V',
                        '--version',
                        action='version',
                        version=pymchelper.__version__)
    group = parser.add_mutually_exclusive_group()
    group.add_argument('-q', '--quiet', action="store_true", help="quiet mode")
    group.add_argument('-v',
                       '--verbosity',
                       action='count',
                       help="increase output verbosity",
                       default=0)
    parsed_args = parser.parse_args(args)

    if parsed_args.verbosity == 1:
        logging.basicConfig(level=logging.INFO)
    elif parsed_args.verbosity > 1:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig()

    cfg = read_config(path=parsed_args.config_path, quiet=parsed_args.quiet)
    t = read_template(cfg=cfg)
    t.write(working_directory=parsed_args.workspace,
            cfg=cfg,
            quiet=parsed_args.quiet)


if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))



from dataclasses import dataclass
import logging
from typing import Optional

import numpy as np

from pymchelper.axis import AxisDescription, MeshAxis
from pymchelper.flair.Input import Particle
from pymchelper.page import Page
from pymchelper.readers.common import ReaderFactory, Reader
from pymchelper.flair.Data import Usrbin, UsrTrack, unpackArray, Usrbdx, Resnuclei, Usrxxx

logger = logging.getLogger(__name__)


class FlukaReaderFactory(ReaderFactory):
    """
    Class responsible for discovery of filetype.
    """

    def get_reader(self):
        """
        Try reading header of Fluka binary file and return a corresponding FlukaReader object

        :return: FlukaReader class if file is digested by Usrxxx Flair reader. None is returned otherwise
        """
        try:
            Usrxxx(self.filename)
            return FlukaReader
        except IOError:
            pass
        return None


class FlukaReader(Reader):

    @property
    def corename(self):
        """
        Fluka output filenames follow this pattern: corename_fort.XX.
        :return: corename part of output file or None in case filename doesn't follow Fluka naming pattern
        """
        core_name = None
        if "_fort" in self.filename:
            core_name = self.filename[-2:]
        return core_name

    def parse_usrbin(self, estimator):
        """
        USRBIN scores distribution of one of several quantities in a regular spatial
        structure (binning detector) independent from the geometry.
        :param estimator: an Estimator object, will be modified here and filled with data
        """
        try:
            usr_object = Usrbin(self.filename)

            # loop over all detectors (pages) in USRBIN object
            for det_no, detector in enumerate(usr_object.detector):
                page = Page(estimator=estimator)
                page.title = detector.name
                # USRBIN doesn't support differential binning type, only spatial binning is allowed

                axes_description = UsrbinScoring.get_axes_description(detector.type)

                estimator.x = MeshAxis(n=detector.nx,
                                       min_val=detector.xlow,
                                       max_val=detector.xhigh,
                                       name=axes_description.x.name,
                                       unit=axes_description.x.unit,
                                       binning=MeshAxis.BinningType.linear)
                estimator.y = MeshAxis(n=detector.ny,
                                       min_val=detector.ylow,
                                       max_val=detector.yhigh,
                                       name=axes_description.y.name,
                                       unit=axes_description.y.unit,
                                       binning=MeshAxis.BinningType.linear)
                estimator.z = MeshAxis(n=detector.nz,
                                       min_val=detector.zlow,
                                       max_val=detector.zhigh,
                                       name=axes_description.z.name,
                                       unit=axes_description.z.unit,
                                       binning=MeshAxis.BinningType.linear)

                # lets check if the detector.score is generalized particle name.
                # if that is the case it means we are scoring Fluence for some particle filter
                # we do a check by querying Flair DB is the particle name is known
                particle_or_scoring_from_id = get_particle_from_db(detector.score)
                if particle_or_scoring_from_id:
                    unit = UsrbinScoring.get_unit_for_scoring(particle_or_scoring_from_id.name)
                    if unit:
                        # here we have the case of genuine scorer (like dose, energy, etc.)
                        page.name = particle_or_scoring_from_id.name
                        page.unit = unit
                    else:
                        # here we have the case of scoring for particles
                        page.name = f"FLUENCE {particle_or_scoring_from_id.name}"
                        page.unit = "/cm^2"
                else:
                    # if not present in the database, we returns the scoring id and empty unit
                    page.name = f"scorer {detector.score}"
                    page.unit = ""

                # unpack detector data
                # TODO cross-check if reshaping is needed
                page.data_raw = np.array(unpackArray(usr_object.readData(det_no)))
                page.error_raw = np.empty_like(page.data_raw)

                estimator.add_page(page)

            return usr_object
        except IOError:
            return None

    def parse_usrbdx(self, estimator):
        """USRBDX defines a detector for a boundary crossing fluence or current estimator.

        :param estimator: an Estimator object, will be modified here and filled with data
        """
        try:
            usr_object = Usrbdx(self.filename)

            # loop over all detectors (pages) in USRBDX object
            for det_no, detector in enumerate(usr_object.detector):
                page = Page(estimator=estimator)
                page.title = detector.name
                page.area = detector.area  # area of the detector in cm**2

                if detector.nb == 1:
                    energy_binning = MeshAxis.BinningType.linear
                    angle_binning = MeshAxis.BinningType.linear
                elif detector.nb == -1:
                    energy_binning = MeshAxis.BinningType.logarithmic
                    angle_binning = MeshAxis.BinningType.linear
                elif detector.nb == 2:
                    energy_binning = MeshAxis.BinningType.linear
                    angle_binning = MeshAxis.BinningType.logarithmic
                elif detector.nb == -2:
                    energy_binning = MeshAxis.BinningType.logarithmic
                    angle_binning = MeshAxis.BinningType.logarithmic
                else:
                    return Exception("Invalid binning type")

                # USRBDX doesn't support spatial (XYZ) binning type
                # USRBDX provides double differential binning, first axis is kinetic energy (in GeV)
                page.diff_axis1 = MeshAxis(
                    n=detector.ne,  # number of energy intervals for scoring
                    min_val=detector.elow,  # minimum kinetic energy for scoring (GeV)
                    max_val=detector.ehigh,  # maximum kinetic energy for scoring (GeV)
                    name="kinetic energy",
                    unit="GeV",
                    binning=energy_binning)

                # second axis is solid angle (in steradians)
                page.diff_axis2 = MeshAxis(
                    n=detector.na,  # number of angular bins
                    min_val=detector.alow,  # minimum solid angle for scoring
                    max_val=detector.ahigh,  # maximum solid angle for scoring
                    name="solid angle",
                    unit="sr",
                    binning=angle_binning)

                # detector.fluence corresponds to i2 in WHAT(1) in first card of USBDX
                if detector.fluence == 1:
                    page.name = "fluence"
                elif detector.fluence == 0:
                    page.name = "current"
                else:
                    page.name = ""
                page.unit = "cm-2 GeV-1 sr-1"

                # TODO If the generalised particle is 208.0 (ENERGY) or 211.0 (EM-ENRGY),
                #             the quantity scored is differential energy fluence (if
                #             cosine-weighted) or differential energy current (energy crossing the
                #             surface). In both cases the quantity will be expressed in GeV per
                #             cm2 per energy unit per steradian per primary

                # unpack detector data
                # TODO cross-check if reshaping is needed
                page.data_raw = np.array(unpackArray(usr_object.readData(det_no)))
                page.error_raw = np.empty_like(page.data_raw)

                estimator.add_page(page)

            return usr_object
        except IOError:
            return None

    def parse_usrtrack(self, estimator):
        """
        :param estimator: an Estimator object, will be modified here and filled with data
        USRTRACK defines a detector for a track-length fluence estimator
        """
        try:
            usr_object = UsrTrack(self.filename)

            # loop over all detectors (pages) in USRTRACK object
            for det_no, detector in enumerate(usr_object.detector):
                page = Page(estimator=estimator)
                page.title = detector.name
                page.volume = detector.volume  # volume of the detector in cm**3

                # USRTRACK doesn't support spatial (XYZ) binning type
                if detector.type == 1:
                    energy_binning = MeshAxis.BinningType.linear
                elif detector.type == -1:
                    energy_binning = MeshAxis.BinningType.logarithmic
                else:
                    return Exception("Invalid binning type")

                # USRTRACK provides single differential binning, with diff axis in kinetic energy (in GeV)
                page.diff_axis1 = MeshAxis(
                    n=detector.ne,  # number of energy intervals for scoring
                    min_val=detector.elow,  # minimum kinetic energy for scoring (GeV)
                    max_val=detector.ehigh,  # maximum kinetic energy for scoring (GeV)
                    name="kinetic energy",
                    unit="GeV",
                    binning=energy_binning)

                page.name = "fluence"
                page.unit = "cm-2 GeV-1"

                # TODO IMPORTANT! The results of USRTRACK are always given as DIFFERENTIAL
                #             distributions of fluence (or tracklength, if the detector region
                #             volume is not specified) in energy, in units of cm-2 GeV-1 (or
                #             cm GeV-1) per incident primary unit weight. Thus, for example, when
                #             requesting a fluence energy spectrum, to obtain INTEGRAL BINNED
                #             results (fluence in cm-2 or tracklength in cm PER ENERGY BIN per
                #             primary) one must multiply the value of each energy bin by the width
                #             of the bin (even for logarithmic binning)

                # TODO If the generalised particle is 208 (ENERGY) or 211 (EM-ENRGY), the
                #             quantity scored is differential energy fluence (or tracklength, if
                #             the detector region volume is not specified), expressed in GeV per
                #             cm2 (or cm GeV) per energy unit per primary. That can sometimes lead
                #             to confusion since GeV cm-2 GeV-1 = cm-2, where energy does not appear.
                #             Note that integrating over energy one gets GeV/cm2.

                # unpack detector data
                # TODO cross-check if reshaping is needed
                page.data_raw = np.array(unpackArray(usr_object.readData(det_no)))
                page.error_raw = np.empty_like(page.data_raw)

                estimator.add_page(page)
            return usr_object
        except IOError:
            return None

    def parse_resnuclei(self, estimator):
        """
        TODO add support for resnuclei
        RESNUCLEi Scores residual nuclei produced in inelastic interactions on a region basis
        :param estimator: an Estimator object, will be modified here and filled with data
        """
        try:
            usr_object = Resnuclei(self.filename)
            # loop over all detectors (pages) in USRTRACK object
            for det_no, detector in enumerate(usr_object.detector):
                page = Page(estimator=estimator)

                # unpack detector data
                # TODO cross-check if reshaping is needed
                page.data_raw = np.array(unpackArray(usr_object.readData(det_no)))
                page.error_raw = np.empty_like(page.data_raw)

                estimator.add_page(page)
            return usr_object
        except IOError:
            return None

    def parse_data(self, estimator):
        """
        TODO
        :param estimator: an Estimator object, will be modified here and filled with data
        """
        for parse_function in (self.parse_usrbin, self.parse_usrtrack, self.parse_usrbdx, self.parse_resnuclei):
            usr_object = parse_function(estimator)

            # stop on first method which return not-None results and return it back
            if usr_object:
                return usr_object

    def read_data(self, estimator, nscale=1):
        """
        TODO
        :param estimator: an Estimator object, will be modified here and filled with data
        """
        usr_object = self.parse_data(estimator)

        estimator.file_counter = usr_object.nbatch
        estimator.number_of_primaries = usr_object.ncase
        estimator.time = usr_object.time
        estimator.title = usr_object.title

        estimator.file_format = 'fluka_binary'
        return True


def get_particle_from_db(particle_id: int) -> Optional[Particle]:
    """Get particle from Flair database by its id"""
    try:
        Particle.makeLists()
        particle = Particle.get(particle_id)
        return particle
    except KeyError:
        return None


@dataclass(frozen=True)
class AxesDescription:
    """Axes descriptions"""

    x: AxisDescription
    y: AxisDescription
    z: AxisDescription


class UsrbinScoring:
    """Helper class for USRBIN scoring"""

    _deposition_scorings = [
        'ENERGY', 'EM-ENRGY', 'DOSE', 'UNB-ENER', 'UNB-EMEN', 'NIEL-DEP', 'DPA-SCO', 'DOSE-EM', 'DOSEQLET', 'RES-NIEL'
    ]
    _fission_density_scorings = ['FISSIONS', 'HE-FISS', 'LE-FISS']
    _neutron_balance_desnity_scorings = ['NEU-BALA']
    _density_of_momentum_scorings = ['X-MOMENT', 'Y-MOMENT', 'Z-MOMENT']
    _activity_scorings = ['ACTIVITY', 'ACTOMASS']
    _dose_equivalent_scorings = ['DOSE-EQ']
    _fluence_weighted_bdf_scorings = ['SI1MEVNE']
    _he_tn_fluence_scorings = ['HEHAD-EQ', 'THNEU-EQ']
    _net_charge_scorings = ['NET-CHRG']

    @classmethod
    def get_unit_for_scoring(cls, scoring: str) -> str:
        """Get unit for scoring name.

        Based on:
        - (1) https://flukafiles.web.cern.ch/manual/chapters/particle_and_material_codes/particles_codes.html
        - (2) https://flukafiles.web.cern.ch/manual/chapters/description_input/description_options/usrbin.html

        :param scoring: scoring name
        :return: tuple of scoring and unit
        """
        if scoring in cls._deposition_scorings:
            if scoring == 'DPA-SCO':
                return '/g'
            if 'DOSE' in scoring:
                return 'GeV/g '
            return 'GeV'
        if scoring in cls._fission_density_scorings:
            return 'fissions/cm^3'
        if scoring in cls._neutron_balance_desnity_scorings:
            return 'neutrons/cm^3'
        if scoring in cls._density_of_momentum_scorings:
            return 'cm^-2'
        if scoring in cls._activity_scorings:
            # This is not totally true, see ACTIVITY and ACTOMASS from 1st link
            if scoring == 'ACTIVITY':
                return 'Bq/cm^3'
            if scoring == 'ACTOMASS':
                return 'Bq/g'
            return ''
        if scoring in cls._dose_equivalent_scorings:
            return 'pSv'
        if scoring in cls._fluence_weighted_bdf_scorings:
            return 'GeV/cm^3'
        if scoring in cls._he_tn_fluence_scorings:
            return 'cm-2'
        if scoring in cls._net_charge_scorings:
            return 'C/cm^3'

        # if unknown scoring, return empty string
        return ''

    @staticmethod
    def get_axes_description(binning_type: int) -> AxesDescription:
        """Get axes descriptions for binning type"""
        if binning_type in (1, 11):  # cylindrical mesh
            return AxesDescription(
                AxisDescription("Radius (R)", "cm"),
                AxisDescription("Angle (PHI)", "rad"),
                AxisDescription("Position (Z)", "cm"))

        # As a default value for not yet implemented binning types.
        # We use cartesian mesh.
        return AxesDescription(
            AxisDescription("Position (X)", "cm"),
            AxisDescription("Position (Y)", "cm"),
            AxisDescription("Position (Z)", "cm"))



#!/usr/bin/env python

import os
import logging
import sys
import argparse
import timeit

from pymchelper.executor.options import SimulationSettings
from pymchelper.executor.runner import OutputDataType, Runner
from pymchelper.simulator_type import SimulatorType
from pymchelper.writers.plots import PlotDataWriter, ImageWriter


def set_logger_level(args):
    """
    Set logger verbosity and quietness based on parsed arguments
    Checks for presence of quiet (-q) and verbose (-v) switches
    """

    # set default logging level to INFO
    level = "INFO"

    # silence almost all warnings if quite option provided
    # there are two levels:
    #  - normal quiet (-q) with level set to WARNING
    #  - very quiet (-qq)  with level set to ERROR
    #  - very very quiet (-qqq and more)  with level set to CRITICAL
    # quiet option is especially useful when using `runmc` commands in the scripts (i.e. in some loops)
    if args.quiet:
        if args.quiet == 1:
            level = "WARNING"
        if args.quiet == 2:
            level = "ERROR"
        else:
            level = "CRITICAL"
    # verbose option, useful in debugging sets DEBUG level
    elif args.verbose:
        level = "DEBUG"

    logging.basicConfig(level=level)


def main(args=None):
    """
    Entry point to the `runmc` script: argument parsing and calling run method
    """

    if args is None:
        args = sys.argv[1:]
    import pymchelper

    parser = argparse.ArgumentParser()
    parser.add_argument('-s', '--simulator', help='Simulator type: shieldhit, topas or fluka '
                                                  '(automatically detected if not provided))',
                        type=str, choices=SimulatorType.names, default=None)
    parser.add_argument('-e', '--executable', help='path to MC executable '
                                                   '(automatically detected if not provided)',
                        type=str, default=None)
    parser.add_argument('-j', '--jobs',
                        help='Number of jobs to run simultaneously (default: {:d})'.format(os.cpu_count()),
                        type=int, default=None)
    parser.add_argument('-m', '--mc-options', help='MC simulation options (default: empty string)',
                        dest='mcopt', type=str, default='')
    parser.add_argument('-o', '--output-dir', help='Output directory (default: .)',
                        dest='outdir', type=str, default='.')
    parser.add_argument('-t', '--out-type', help='output data type (default {:s})'.format(OutputDataType.txt.name),
                        dest='outtype', type=str, nargs='*',
                        choices=[x.name for x in OutputDataType], default=OutputDataType.txt.name)
    parser.add_argument('-w', '--work-dir', help='Workspace directory (default: .)',
                        dest='workspace', type=str, default='.')
    parser.add_argument('-k', '--keep', action='store_true', help='keep workspace directories')
    parser.add_argument('-q', '--quiet', action='count', default=0, help='be silent')
    parser.add_argument('-v',
                        '--verbose',
                        action='count',
                        default=0,
                        help='give more output. Option is additive, and can be used up to 3 times')
    parser.add_argument('-V', '--version', action='version', version=pymchelper.__version__)
    parser.add_argument('input', help='input filename or directory', type=str)

    parsed_args = parser.parse_args(args)

    # set verbose and quietness options
    set_logger_level(parsed_args)

    # strip MC simulation arguments:
    #   we have possibility to pass extra options to MS simulation executable
    #   passing these options in most obvious way, i.e. -m --time 00:15:30 won't work
    #   as argument parsing library will interpret --time as runmc option and not as -m option value
    #   surrounding --time 00:15:30 with "" won't help, as they could be stripped away by shell (i.e. bash)
    #   the only possible way is to embed the MC simulator options with custom wrapping characters, like [,]:
    #      -m "[--time 00:15:30 -v -n 1000]"
    #   here we strip these wrapping characters, if present
    # if no -m option is provided then we need to deal with empty string
    parsed_simulation_opts = parsed_args.mcopt
    # check if list is not None, and if it has at least one element
    if parsed_simulation_opts:
        # check if parsed options are embedded in [,]
        if parsed_simulation_opts[0] == '[' and parsed_simulation_opts[-1] == ']':
            parsed_simulation_opts = parsed_simulation_opts[1:-1]  # strip the list from surrounding brackets

    # set MC simulation settings based on:
    #   - MC simulation input file (i.e. *.inp file for FLUKA) or
    #     directories (i.e. directory with beam.dat, geo.dat etc for SHIELD-HIT12A)
    #   - location of MC simulator executable file (i.e. `shieldhit` or `rfluka`)
    #   - simulation options for the MC engine provided via -m switch (i.e. --time or -v)
    settings = SimulationSettings(input_path=parsed_args.input,
                                  simulator_type=SimulatorType.from_name(parsed_args.simulator),
                                  simulator_exec_path=parsed_args.executable,
                                  cmdline_opts=parsed_simulation_opts)

    # create runner object based on MC options and dedicated parallel jobs number
    # note that runner object is only created here, no simulation is started at this point
    # and no directories are being created
    runner_obj = Runner(settings=settings,
                        jobs=parsed_args.jobs,
                        keep_workspace_after_run=parsed_args.keep,
                        output_directory=parsed_args.outdir)

    # start parallel execution of MC simulation
    # temporary directories needed for parallel execution as well as the output are being saved in `outdir`
    # in case of successful execution this would return list of temporary workspaces directories
    # containing partial results from simultaneous parallel executions
    start_time = timeit.default_timer()
    runner_obj.run()
    elapsed = timeit.default_timer() - start_time
    print("MC simulation took {:.3f} seconds".format(elapsed))

    # if simulation was successful proceed to data extraction by combining partial results from simultaneous executions
    # each simulation can produce multiple files
    # results are stored in a dictionary (`data_dict`) with keys being filenames
    # and values being pymchelper `Estimator` objects (which keep i.e. numpy arrays with results)
    data_dict = runner_obj.get_data()

    # if user requests combined results as text files, the code below is used to convert Estimator objects to them
    # note that multiple text files can be created here
    start_time = timeit.default_timer()
    if data_dict and (OutputDataType.txt.name in parsed_args.outtype):
        for core_filename in data_dict:
            logging.debug("Core filename {:s}".format(core_filename))
            output_file = os.path.join(parsed_args.outdir, core_filename)
            writer = PlotDataWriter(output_file, None)
            writer.write(data_dict[core_filename])

    # if user requests combined results as PNG image, the code below is used to convert Estimator objects to them
    # note that multiple PNG files can be created here
    if data_dict and (OutputDataType.plot.name in parsed_args.outtype):
        for core_filename in data_dict:
            output_file = os.path.join(parsed_args.outdir, core_filename)
            writer = ImageWriter(output_file, argparse.Namespace(colormap='gnuplot2', log=''))
            writer.write(data_dict[core_filename])

    elapsed = timeit.default_timer() - start_time
    print("Saving output {:.3f} seconds".format(elapsed))

    runner_obj.clean()

    return 0


if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))



PROC_MIN_FIELD_WIDTH = 46
PROC_DECIMAL_CHAR = '.'
PROC_NO_LEADING_BLANK = False

# based on https://bitbucket.org/brendanarnold/py-fortranformat


def _swapchar(s, ind, newch):
    """
    Helper function to make chars in a string mutableish
    """
    if 0 < ind >= len(s):
        raise IndexError('index out of range')
    return s[:ind] + newch + s[ind + 1:]


def _compose_nan_string(w, ftype):
    if ftype in ['B', 'O', 'Z']:
        return ''
    else:
        # Allow at least 'NaN' to be printed
        if w == 0:
            w = 4  # n.b. this is what is set in Gfortran 4.4.0
        if w < 3:
            return '*' * w
    return 'NaN'.rjust(w)


def _calculate_sign(state, negative_flag):
    s = ''
    if negative_flag:
        s = '-'
    elif state['incl_plus']:
        s = '+'
    else:
        s = ''
    return s


def _compose_inf_string(w, ftype, sign_bit):
    if ftype in ['B', 'O', 'Z']:
        return ''
    else:
        sign = '+'
        # Allow at least 'Inf' to be printed
        if w == 0:
            w = 4
        if w < 3:
            return '*' * w
        # Change sign if negative
        if sign_bit:
            sign = '-'
            # Require sign if negative, if no space then overflow
            if w == 3:
                return '*' * w
        # Output long version if long enough
        if w > 8:
            return (sign + 'Infinity').rjust(w)
        # Output shortened version with sign if long enough
        elif w > 3:
            return (sign + 'Inf').rjust(w)
        # Should only output short version with no sign if positive
        else:
            return 'Inf'


def _compose_float_string(w, e, d, state, val, ftype):
    """
    Adapted from code in glibfortran which is written in C so is somwhat
    'bit-pushy' in nature. Writes the value to an initial string (buffer)
    and then pulls the subsequent strings from that
    """
    if (d < 0) or (d is None):
        raise Exception('Unspecified precision')
    # Make sure they are ints
    d = int(round(d))
    if e is not None:
        e = int(round(e))
    if w is not None:
        w = int(round(w))
    # ==== write_float ==== (function)
    edigits = 4  # Largest number of exponent digits expected
    # Otherwise convert knowing what the required precision is (i.e. knowing d)
    ndigits = d
    if ndigits > (PROC_MIN_FIELD_WIDTH - 4 - edigits):
        ndigits = PROC_MIN_FIELD_WIDTH - 4 - edigits
    # ==== WRITE_FLOAT ==== (macro)
    # Determine sign of value
    if val == 0.0:
        sign_bit = '-' in str(val)
    else:
        sign_bit = val < 0
    # handle the nan and inf cases
    if type(val) is float and val != val:
        return _compose_nan_string(w, ftype)
    infinity = 1e1000000
    if val in (-infinity, infinity):
        return _compose_inf_string(w, ftype, sign_bit)
    tmp = abs(val)
    # Round the input if the input is less than 1
    zero_flag = (tmp == 0)
    # === DTOA === (macro)
    # write the tmp value to the string buffer
    # sprintf seems to allow negative number of decimal places,
    # need to correct for this
    if ndigits <= 0:
        fmt = "%+-#{:d}e".format(PROC_MIN_FIELD_WIDTH)
    else:
        fmt = "%+-#{:d}.{:d}e".format(PROC_MIN_FIELD_WIDTH, ndigits - 1)
    buff = fmt % tmp
    # === WRITE_FLOAT === (macro)
    return _output_float(w, d, e, state, ftype, buff, sign_bit, zero_flag, ndigits, edigits)


def _output_float(w, d, e, state, ft, buff, sign_bit, zero_flag, ndigits, edigits):
    """
    TODO
    :param w:
    :param d:
    :param e:
    :param state:
    :param ft:
    :param buff:
    :param sign_bit:
    :param zero_flag:
    :param ndigits:
    :param edigits:
    :return:
    """
    # nbefore - number of digits before the decimal point
    # nzero - number of zeros after the decimal point
    # nafter - number of digits after the decimal point
    # nzero_real - number of zeros after the decimal point
    #               regardles of the precision

    # Some hacks to change None to -1 (C convention)
    if w is None:
        w = -1
    if e is None:
        e = -1
    nzero_real = -1
    sign = _calculate_sign(state, sign_bit)
    # Some debug
    if d != 0:
        assert (buff[2] in ['.', ','])
        assert (buff[ndigits + 2] == 'e')
    # Read in the exponent
    ex = int(buff[ndigits + 3:]) + 1
    # Handle zero case
    if zero_flag:
        ex = 0
        sign = _calculate_sign(state, False)
        # Handle special case
        if w == 0:
            w = d + 2
    # Get rid of the decimal and the initial sign i.e. normalise the digits
    digits = buff[1] + buff[3:]
    # Find out where to place the decimal point
    if ft in ['E', 'D']:
        i = state['scale']
        if (d <= 0) and (i == 0):
            raise Exception("Precision not greater " "than zero in format specifier 'E' or 'D'")
        if (i <= -d) or (i >= (d + 2)):
            raise Exception("Scale factor out of range " "in format specifier 'E' or 'D'")
        if not zero_flag:
            ex = ex - i
        if i < 0:
            nbefore = 0
            nzero = -i
            nafter = d + i
        elif i > 0:
            nbefore = i
            nzero = 0
            nafter = (d - i) + 1
        else:
            nbefore = 0
            nzero = 0
            nafter = d
        expchar = ft
    # Round the value
    if (nbefore + nafter) == 0:
        ndigits = 0
        if (nzero_real == d) and (int(digits[0]) >= 5):
            # We rounded to zero but shouldn't have
            nzero -= 1
            nafter = 1
            digits = '1' + digits[1:]
            ndigits = 1
    elif (nbefore + nafter) < ndigits:
        ndigits = nbefore + nafter
        i = ndigits
        if int(digits[i]) >= 5:
            # Propagate the carry
            i -= 1
            while i >= 0:
                digit = int(digits[i])
                if digit != 9:
                    digits = _swapchar(digits, i, str(digit + 1))
                    break
                else:
                    digits = _swapchar(digits, i, '0')
                i -= 1
            # Did the carry overflow?
            if i < 0:
                digits = '1' + digits
                if ft == 'F':
                    if nzero > 0:
                        nzero -= 1
                        nafter += 1
                    else:
                        nbefore += 1
                elif ft == 'EN':
                    nbefore += 1
                    if nbefore == 4:
                        nbefore = 1
                        ex += 3
                else:
                    ex += 1
    # Calculate the format of the exponent field
    if expchar is not None:
        # i = abs(ex)
        # while i >= 10:
        #     edigits = edigits + 1
        #     i = i / 10.0
        if e < 0:
            # Width not specified, must be no more than 3 digits
            if (ex > 999) or (ex < -999):
                edigits = -1
            else:
                edigits = 4
                if (ex > 99) or (ex < -99):
                    expchar = ' '
        else:
            assert (isinstance(ex, int))
            edigits = len(str(abs(ex)))
            # Exponenet width specified, check it is wide enough
            if edigits > e:
                edigits = -1
            else:
                edigits = e + 2
    else:
        edigits = 0
    # Zero values always output as positive,
    # even if the value was egative before rounding
    i = 0
    while i < ndigits:
        if digits[i] != '0':
            break
        i += 1
    if i == ndigits:
        # The output is zero so set sign accordingly
        sign = _calculate_sign(state, False)
    # Pick a field size if none was specified
    if w <= 0:
        w = nbefore + nzero + nafter + 1 + len(sign)
    # Work out how much padding is needed
    nblanks = w - (nbefore + nzero + nafter + edigits + 1)
    if sign != '':
        nblanks -= 1
    # Check value fits in specified width
    if (nblanks < 0) or (edigits == -1):
        return '*' * w
    # See if we have space for a zero before the decimal point
    if (nbefore == 0) and (nblanks > 0):
        leadzero = True
        nblanks -= 1
    else:
        leadzero = False
    out = ''
    # Pad to full field width
    if nblanks > 0:  # dtp->u.p.no_leading_blank
        out += ' ' * nblanks
    # Attach the sign
    out += sign
    # Add the lead zero if necessary
    if leadzero:
        out += '0'
    # Output portion before the decimal point padded with zeros
    if nbefore > 0:
        if nbefore > ndigits:
            out = out + digits[:ndigits] + (' ' * (nbefore - ndigits))
            digits = digits[ndigits:]
            ndigits = 0
        else:
            i = nbefore
            out += digits[:i]
            digits = digits[i:]
            ndigits -= i
    # Output the decimal point
    out += PROC_DECIMAL_CHAR
    # Output the leading zeros after the decimal point
    if nzero > 0:
        out += '0' * nzero
    # Output the digits after the decimal point, padded with zeros
    if nafter > 0:
        if nafter > ndigits:
            i = ndigits
        else:
            i = nafter
        zeros = '0' * (nafter - i)
        out = out + digits[:i] + zeros
        digits = digits[nafter:]
        ndigits -= nafter
    # Output the exponent
    if expchar is not None:
        if expchar != ' ':
            out += expchar
            edigits -= 1
        fmt = '%+0' + str(edigits) + 'd'
        tmp_buff = fmt % ex
        out += tmp_buff
    return out


def format_d(w, d, val):
    """
    TODO
    :param w:
    :param d:
    :param val:
    :return:
    """
    e = None
    ftype = 'D'
    state = {'blanks_as_zeros': False, 'incl_plus': False, 'position': 0, 'scale': 0, 'halt_if_no_vals': False}
    return _compose_float_string(w, e, d, state, val, ftype)


def format_e(w, d, val):
    """
    TODO
    :param w:
    :param d:
    :param val:
    :return:
    """
    e = None
    ftype = 'E'
    state = {'blanks_as_zeros': False, 'incl_plus': False, 'position': 0, 'scale': 0, 'halt_if_no_vals': False}
    return _compose_float_string(w, e, d, state, val, ftype)



import logging
import json

from pymchelper.axis import MeshAxis
from pymchelper.estimator import Estimator
from pymchelper.page import Page


logger = logging.getLogger(__name__)


class JsonWriter:
    """
    Supports writing JSON format.
    JSON format is a format accepted by yaptide project.
    """

    def __init__(self, filename, options):
        self.filename = filename
        self.options = options
        if not self.filename.endswith(".json"):
            self.filename += ".json"

    def write(self, estimator: Estimator):
        """Writes estimator object to json file"""
        if len(estimator.pages) == 0:
            print("No pages in the output file, conversion to JSON skipped.")
            return False

        est_dict = {
            "metadata": {},
            "pages": []
        }
        exclude = {"data_raw", "error_raw", "estimator", "diff_axis1", "diff_axis2"}
        exclude |= set(estimator.__dict__.keys())

        # read metadata from estimator object
        for name, value in estimator.__dict__.items():
            # skip non-metadata fields
            if name not in {"data", "data_raw", "error", "error_raw", "counter", "pages", "x", "y", "z"}:
                # remove \" to properly generate JSON
                est_dict["metadata"][name] = str(value).replace("\"", "")

        for page in estimator.pages:
            # page_dict contains:
            # "dimensions" indicating it is 1 dim page
            # "data" which has unit, name and list of data values
            page: Page
            page_dict = {
                "metadata": {},
                "dimensions": page.dimension,
                "data": {
                    "unit": str(page.unit),
                    "name": str(page.name),
                }
            }

            # read metadata from page object
            for name, value in page.__dict__.items():
                # skip non-metadata fields and fields already read from estimator object
                if name not in exclude:
                    # remove \" to properly generate JSON
                    page_dict["metadata"][name] = str(value).replace("\"", "")

            if page.dimension == 0:
                page_dict["data"]["values"] = [page.data_raw.tolist()]
            else:
                page_dict["data"]["values"] = page.data_raw.tolist()

            for i in range(page.dimension):
                axis: MeshAxis = page.plot_axis(i)
                page_dict[f"axis_dim{i+1}"] = {
                    "unit": str(axis.unit),
                    "name": str(axis.name),
                    "values": axis.data.tolist(),
                }

            est_dict["pages"].append(page_dict)

        with open(self.filename, "w") as json_file:
            json.dump(est_dict, json_file)

        return 0



import numpy as np

from pymchelper.axis import MeshAxis, AxisId
from pymchelper.shieldhit.detector.detector_type import SHDetType


class Page:
    def __init__(self, estimator=None):

        self.estimator = estimator

        self.data_raw = np.array([float("NaN")])  # linear data storage
        self.error_raw = np.array([float("NaN")])  # linear data storage

        self.name : str = ""
        self.unit : str = ""

        self.dettyp = None  # Dose, Fluence, LET etc...

        # optional first differential axis
        self.diff_axis1 = MeshAxis(n=1,
                                   min_val=float("NaN"),
                                   max_val=float("NaN"),
                                   name="",
                                   unit="",
                                   binning=MeshAxis.BinningType.linear)

        # optional second differential axis
        self.diff_axis2 = MeshAxis(n=1,
                                   min_val=float("NaN"),
                                   max_val=float("NaN"),
                                   name="",
                                   unit="",
                                   binning=MeshAxis.BinningType.linear)

    def axis(self, axis_id):
        """
        TODO
        """
        if axis_id == AxisId.diff1:
            return self.diff_axis1
        elif axis_id == AxisId.diff2:
            return self.diff_axis2
        elif self.estimator:
            return self.estimator.axis(axis_id)
        return None

    @property
    def dimension(self) -> int:
        """
        Let's take again detector d with YZ scoring.
        >>> from pymchelper.estimator import Estimator
        >>> e = Estimator()
        >>> e.x = MeshAxis(n=1, min_val=0.0, max_val=1.0, name="X", unit="cm", binning=MeshAxis.BinningType.linear)
        >>> e.y = MeshAxis(n=3, min_val=0.0, max_val=150.0, name="Y", unit="cm", binning=MeshAxis.BinningType.linear)
        >>> e.z = MeshAxis(n=2, min_val=0.0, max_val=2.0, name="Z", unit="cm", binning=MeshAxis.BinningType.linear)
        >>> e.dimension
        2
        >>> p = Page(e)
        >>> p.diff_axis1 = MeshAxis(n=10, min_val=0.0, max_val=100.0, name="E", unit="MeV",
        ...                         binning=MeshAxis.BinningType.linear)
        >>> p.dimension
        3

        :return: number of page axes (including differential) which have more than one bin
        """
        if self.estimator:
            return 2 - (self.diff_axis1.n, self.diff_axis2.n).count(1) + self.estimator.dimension
        else:
            return 0

    @property
    def data(self):
        """
        3-D view of page data.

        Page data is stored originally in `data_raw` 1-D array.
        This property provides efficient view of data, suitable for numpy-like indexing.

        >>> from pymchelper.estimator import Estimator
        >>> e = Estimator()
        >>> e.x = MeshAxis(n=2, min_val=0.0, max_val=10.0, name="X", unit="cm", binning=MeshAxis.BinningType.linear)
        >>> e.y = MeshAxis(n=3, min_val=0.0, max_val=150.0, name="Y", unit="cm", binning=MeshAxis.BinningType.linear)
        >>> e.z = MeshAxis(n=1, min_val=0.0, max_val=1.0, name="Z", unit="cm", binning=MeshAxis.BinningType.linear)
        >>> p = Page(estimator=e)
        >>> p.data_raw = np.arange(6)
        >>> tuple(int(n) for n in p.data.shape)
        (2, 3, 1, 1, 1)
        >>> p.data[1, 2, 0, 0, 0]
        5

        :return: reshaped view of ``data_raw``
        """
        if self.estimator:
            # phase space data needs to be reshaped to a 2D array
            if self.dettyp == SHDetType.mcpl:
                return self._reshape(data_1d=self.data_raw, shape=(8, -1))
            return self._reshape(data_1d=self.data_raw,
                                 shape=(self.estimator.x.n, self.estimator.y.n, self.estimator.z.n,
                                        self.diff_axis1.n, self.diff_axis2.n))
        return self.data_raw

    @property
    def error(self):
        """
        3-D view of page error

        For more details see ``data`` property.
        :return:
        """
        if self.estimator:
            return self._reshape(data_1d=self.error_raw,
                                 shape=(self.estimator.x.n, self.estimator.y.n, self.estimator.z.n,
                                        self.diff_axis1.n, self.diff_axis2.n))
        return self.error_raw

    def _reshape(self, data_1d, shape : tuple):
        # TODO check also  tests/res/shieldhit/single/ex_yzmsh.bdo as it is saved in bin2010 format
        if self.estimator:
            order = 'C'
            if self.estimator.file_format in {'bdo2016', 'bdo2019', 'fluka_binary'}:
                order = 'F'
            return data_1d.reshape(shape, order=order)
        else:
            return data_1d

    def plot_axis(self, id : int):
        """
        Calculate new order of detector axis, axis with data (n>1) comes first
        Axes with constant value goes last.

        Let's take a detector d with YZ scoring.
        >>> from pymchelper.estimator import Estimator
        >>> e = Estimator()
        >>> e.x = MeshAxis(n=1, min_val=0.0, max_val=1.0, name="X", unit="cm", binning=MeshAxis.BinningType.linear)
        >>> e.y = MeshAxis(n=3, min_val=0.0, max_val=150.0, name="Y", unit="cm", binning=MeshAxis.BinningType.linear)
        >>> e.z = MeshAxis(n=2, min_val=0.0, max_val=2.0, name="Z", unit="cm", binning=MeshAxis.BinningType.linear)
        >>> p = Page(estimator=e)
        >>> e.add_page(page=p)

        First axis for plotting will be Y (as X axis holds only one bin):
        >>> p.plot_axis(0)
        MeshAxis(n=3, min_val=0.0, max_val=150.0, name='Y', unit='cm', binning=<BinningType.linear: 0>)

        Second axis for plotting will be Z (its the next after Y with n > 1 bins)
        >>> p.plot_axis(1)
        MeshAxis(n=2, min_val=0.0, max_val=2.0, name='Z', unit='cm', binning=<BinningType.linear: 0>)

        Finally the third axis will be X, but it cannot be used for plotting as it has only one bin.
        >>> p.plot_axis(2)
        MeshAxis(n=1, min_val=0.0, max_val=1.0, name='X', unit='cm', binning=<BinningType.linear: 0>)


        :param id: axis number (0, 1, 2, 3 or 4)
        :return: axis object
        """
        plotting_order = (AxisId.x, AxisId.y, AxisId.z, AxisId.diff1, AxisId.diff2)
        variable_axes_id = [i for i in plotting_order if self.axis(i).n > 1]
        constant_axes_id = [i for i in plotting_order if self.axis(i).n == 1]
        plotting_order = variable_axes_id + constant_axes_id
        return self.axis(plotting_order[id])



from pathlib import Path
import re
from typing import List, Optional, Tuple

import numpy as np
from pymchelper.axis import MeshAxis
from pymchelper.estimator import Estimator
from pymchelper.page import Page
from pymchelper.readers.common import Reader, ReaderFactory


class TopasReaderFactory(ReaderFactory):
    """Factory for TopasReader"""

    def get_reader(self):
        """Return TopasReader if the file extension is .csv"""
        if self.filename.endswith('.csv'):
            return TopasReader
        return None


def extract_parameter_filename(header_line: str) -> Optional[str]:
    """Get parameter filename from the output file"""
    pattern = r"# Parameter File: (.*)"
    match = re.search(pattern, header_line)
    parameter_filename = None
    if match:
        parameter_filename = match.group(1)
    return parameter_filename


def extract_bins_data(dimensions: List[str], header_lines: List[str]) -> Optional[dict]:
    """
    Takes as arguments a list of dimensions (e.g. ['X', 'Y', 'Z'])
    and three consecutive lines from the output file header.
    Returns dict containing number of bins, bin size and unit for each dimension
    or None if output file does not contain this information for provided dimensions
    """
    if len(dimensions) != 3 or len(header_lines) != 3:
        return None
    bins_data = {}
    pattern = r"# {} in (\d+) bin[s ] of ([\d.]+) (\w+)"
    for line_index, dimension in enumerate(dimensions):
        match = re.search(pattern.format(dimension), header_lines[line_index])
        if match:
            bins_data[dimension] = {'num': int(match.group(1)),
                                    'size': float(match.group(2)),
                                    'unit': match.group(3)}
        else:
            return None
    return bins_data


def extract_scorer_name(header_line: str) -> Optional[str]:
    """Get scorer name from the output file"""
    pattern = r"# Results for scorer: (\w+)"
    match = re.search(pattern, header_line)
    name = None
    if match:
        name = match.group(1)
    return name


def extract_scorer_unit_results(header_line: str) -> Optional[Tuple[str, str, List]]:
    """Get scoring quantity, unit and the scoring values (sum/mean/etc.) from the output file"""
    scorers = ['DoseToMedium', 'DoseToWater', 'DoseToMaterial', 'TrackLengthEstimator',
               'AmbientDoseEquivalent', 'EnergyDeposit', 'Fluence', 'EnergyFluence',
               'StepCount', 'OpticalPhotonCount', 'OriginCount', 'Charge', 'EffectiveCharge',
               'ProtonLET', 'SurfaceCurrent', 'SurfaceTrackCount', 'PhaseSpace']
    for scorer in scorers:
        if scorer in header_line:
            unit = ""
            pattern = f"# {scorer}\\s(\\( .*? \\))?\\s*: (.*)"
            match = re.search(pattern, header_line)
            if match:
                unit = ""
                if match.group(1):
                    unit = match.group(1)[2:-2]
                results = match.group(2).split()
                return (scorer, unit, results)
    return None


def extract_differential_axis(header_line: str) -> Optional[MeshAxis]:
    """Check if the output file contains differential axis and get it from file if it does"""
    if "# Binned by" in header_line:
        pattern = r"# Binned by (.+?) in (\d+) bin[s ] of (\d+) (\w+) from ([\d.]+) (\w+) to ([\d.]+) (\w+)"
        match = re.search(pattern, header_line)
        if match:
            binned_by = match.group(1)
            num_bins = int(match.group(2))
            unit = match.group(4)
            min_val = float(match.group(5))
            max_val = float(match.group(7))

            return MeshAxis(n=num_bins, min_val=min_val, max_val=max_val,
                            name=binned_by, unit=unit, binning=MeshAxis.BinningType.linear)
    return None


class TopasReader(Reader):
    """Reader for Topas output files"""

    def __init__(self, filename):
        super(TopasReader, self).__init__(filename)
        self.directory = Path(filename).parent

    def read_data(self, estimator: Estimator) -> bool:
        """
        Read the data from the file and store them in the provided estimator object.
        TOPAS reader assumes that the input file is in the same directory as the output file
        to extract the number of histories from it.
        If the name of the input file is not found in the output file
        or the input file is not in the same directory, the number of histories is set to 0.
        """
        with open(self.filename, 'r') as results_file:
            results_lines = results_file.readlines()
            header_lines = [line for line in results_lines if line.startswith("#")]

            num_histories = 0
            for line in header_lines:
                input_filename = extract_parameter_filename(line)
                if input_filename is not None:
                    input_file_path = Path(self.directory) / input_filename
                    if input_file_path.exists():
                        with open(input_file_path, 'r') as input_file:
                            pattern = r'NumberOfHistoriesInRun\s*=\s*(\d+)'
                            input_data = input_file.read()
                            match = re.search(pattern, input_data)
                            if match:
                                number_str = re.search(r'\d+', match.group())
                                if number_str:
                                    num_histories = int(number_str.group())
                    break

            estimator.file_corename = Path(self.filename).name[:-4]
            estimator.number_of_primaries = num_histories
            estimator.file_format = "csv"

            dimensions = [['X', 'Y', 'Z'],
                          ['R', 'Phi', 'Z'],
                          ['R', 'Phi', 'Theta']]

            actual_dimensions = None
            for curr_dimensions in dimensions:
                for line_index in range(len(header_lines)-3):
                    bins_data = extract_bins_data(curr_dimensions, header_lines[line_index:line_index+3])
                    if bins_data is not None:
                        actual_dimensions = curr_dimensions
                        x_max = bins_data[actual_dimensions[0]]['size']*bins_data[actual_dimensions[0]]['num']
                        estimator.x = MeshAxis(n=bins_data[actual_dimensions[0]]['num'],
                                               min_val=0.0, max_val=x_max,
                                               name=actual_dimensions[0], unit=bins_data[actual_dimensions[0]]['unit'],
                                               binning=MeshAxis.BinningType.linear)
                        y_max = bins_data[actual_dimensions[1]]['size']*bins_data[actual_dimensions[1]]['num']
                        estimator.y = MeshAxis(n=bins_data[actual_dimensions[1]]['num'],
                                               min_val=0.0, max_val=y_max,
                                               name=actual_dimensions[1], unit=bins_data[actual_dimensions[1]]['unit'],
                                               binning=MeshAxis.BinningType.linear)
                        z_max = bins_data[actual_dimensions[2]]['size']*bins_data[actual_dimensions[2]]['num']
                        estimator.z = MeshAxis(n=bins_data[actual_dimensions[2]]['num'],
                                               min_val=0.0, max_val=z_max,
                                               name=actual_dimensions[2], unit=bins_data[actual_dimensions[2]]['unit'],
                                               binning=MeshAxis.BinningType.linear)
                        no_bins = False
                        break
                if actual_dimensions is not None:
                    break

            if bins_data is None:
                # We assume that the geometry is not divided into bins
                # (or in other words there is one bin with one score)
                no_bins = True

            for line in header_lines:
                differential_axis = extract_differential_axis(line)
                if differential_axis is not None:
                    break

            # In one output csv file there can be multiple results for one scorer
            # (e.g. mean and standard deviation for fluence).
            # We look for mean and standard deviation and ignore the rest.
            # Then we put them in raw_data and raw_error respectively.
            scorer, unit, results = None, None, None
            for line in header_lines:
                res = extract_scorer_unit_results(line)
                if res is not None:
                    scorer, unit, results = res
                    break

            # We did not find scorer info in the output file
            if scorer is None:
                return False

            num_results = len(results)
            page = Page(estimator=estimator)
            set_data = False
            set_error = False
            for column, result in enumerate(results):
                if result not in ['Mean', 'Standard_Deviation']:
                    continue
                if differential_axis:
                    page.diff_axis1 = differential_axis
                    # When there is a differential axis, each line in csv contains scores
                    # in alternating order. For example, for sum and mean it looks like this:
                    # sum(underflow) mean(underflow) sum(bin1) mean(bin1) sum(bin2) mean(bin2) ...
                    # Binning by time has one additional bin at the end : overflow.
                    # Binning by energy has two additional bins at the end:
                    # overflow and for case of no incident track.
                    # Both have one additional bin at the beginning: underflow.
                    # We ignore these additional bins.
                    if differential_axis.name == 'time':
                        additional_bins = 1
                    elif differential_axis.name == 'incident track energy':
                        additional_bins = 2
                    else:
                        return False
                    lines = np.genfromtxt(self.filename, delimiter=',')
                    last_bin_index = len(lines[0]) - num_results*additional_bins
                    scores = lines[:, column+num_results:last_bin_index:num_results].flatten()

                else:
                    # When there is no differential axis, each line in csv file looks like this:
                    # x y z result1 result2 result3 ...
                    # where x, y, z are coordinates of the bin - we ignore them.
                    # Unless there are no bins - then there is one line with scores
                    # result1 result2 ...
                    if no_bins:
                        data = np.genfromtxt(self.filename, delimiter=',')
                        if data.shape == ():
                            scores = np.array([data])
                        else:
                            scores = np.array([data[column]])
                    else:
                        lines = np.genfromtxt(self.filename, delimiter=',')
                        scores = lines[:, column+3]

                for line in header_lines:
                    title = extract_scorer_name(line)
                    if title is not None:
                        page.title = title
                        page.name = title
                        break

                page.dettyp, page.unit = scorer, unit

                if result == 'Mean':
                    page.data_raw = scores
                    set_data = True
                elif result == 'Standard_Deviation':
                    page.error_raw = scores
                    set_error = True

            # If we didn't find mean results for the scorer, we return False
            if not set_data:
                return False
            if not set_error:
                page.error_raw = np.empty_like(page.data_raw)
            estimator.add_page(page)
            return True

    @property
    def corename(self) -> str:
        return self.filename[:-4]



from collections import namedtuple
from dataclasses import dataclass
from enum import IntEnum

import numpy as np


class MeshAxis(namedtuple('MeshAxis', 'n min_val max_val name unit binning')):
    """
    Scoring mesh axis data.

    It can represent an axis variety of scorers:
    x,y or z in cartesian scoring, r, rho or z in cylindrical.
    An axis represents a sequence of ``n`` numbers, defining linear or logarithmic binning.
    This sequence of numbers is not stored in the memory, but can be generated using data property method.
    ``min_val`` is lowest bin left edge, max_val is highest bin right edge
    ``name`` can be used to define physical quantity (i.e. position, energy, angle).
    ``unit`` gives physical units (i.e. cm, MeV, mrad).

    ``MeshAxis`` is constructed as immutable data structure, thus it is possible
    to set field values only upon object creation. Later they are available for read only.

    >>> x = MeshAxis(n=10, min_val=0.0, max_val=30.0, name="Position", unit="cm", binning=MeshAxis.BinningType.linear)
    >>> x.n, x.min_val, x.max_val
    (10, 0.0, 30.0)
    >>> x.n = 5
    Traceback (most recent call last):
    ...
    AttributeError: can't set attribute

    ``binning`` field (use internal ``BinningType.linear`` or ``BinningType.logarithmic``) can distinguish
    log from linear binning
    """
    class BinningType(IntEnum):
        """
        type of axis generator
        """
        linear = 0
        logarithmic = 1

    @property
    def data(self):
        """
        Generates linear or logarithmic sequence of ``n`` numbers.

        These numbers are middle points of the bins
        defined by ``n``, ``min_val`` and ``max_val`` parameters.

        >>> x = MeshAxis(n=10, min_val=0.0, max_val=10.0, name="X", unit="cm", binning=MeshAxis.BinningType.linear)
        >>> x.data
        array([ 0.5,  1.5,  2.5,  3.5,  4.5,  5.5,  6.5,  7.5,  8.5,  9.5])

        Binning may also consist of one bin:
        >>> x = MeshAxis(n=1, min_val=0.0, max_val=5.0, name="X", unit="cm", binning=MeshAxis.BinningType.linear)
        >>> x.data
        array([ 2.5])

        Logarithmic binning works as well, middle bin points are calculated as geometrical mean.
        Here we define 3 bins: [1,4], [4,16], [16,64].
        >>> x = MeshAxis(n=3, min_val=1.0, max_val=64.0, name="X", unit="cm", binning=MeshAxis.BinningType.logarithmic)
        >>> x.data
        array([  2.,   8.,  32.])

        For the same settings as below linear scale gives as expected different sequence:
        >>> x = MeshAxis(n=3, min_val=1.0, max_val=64.0, name="X", unit="cm", binning=MeshAxis.BinningType.linear)
        >>> x.data
        array([ 11.5,  32.5,  53.5])

        For logarithmic axis min_val has to be positive:
        >>> x = MeshAxis(n=3, min_val=-2.0, max_val=64.0, name="X", unit="cm", binning=MeshAxis.BinningType.logarithmic)
        >>> x.data
        Traceback (most recent call last):
        ...
        Exception: Left edge of first bin (-2) is not positive

        :return:
        """
        if self.max_val < self.min_val:
            raise Exception("Right edge of last bin ({:g}) is smaller than left edge of first bin ({:g})".format(
                self.max_val, self.min_val
            ))
        if self.binning == self.BinningType.linear:
            bin_width = (self.max_val - self.min_val) / self.n
            first_bin_mid = self.min_val + bin_width / 2.0  # arithmetic mean
            last_bin_mid = self.max_val - bin_width / 2.0  # arithmetic mean
            return np.linspace(start=first_bin_mid, stop=last_bin_mid, num=self.n)
        elif self.binning == self.BinningType.logarithmic:
            if self.min_val < 0.0:
                raise Exception("Left edge of first bin ({:g}) is not positive".format(self.min_val))
            q = (self.max_val / self.min_val)**(1.0 / self.n)  # an = a0 q^n
            first_bin_mid = self.min_val * q**0.5  # geometrical mean sqrt(a0 a1) = sqrt( a0 a0 q) = a0 sqrt(q)
            last_bin_mid = self.max_val / q**0.5  # geometrical mean sqrt(an a(n-1)) = sqrt( an an/q) = an / sqrt(q)
            try:
                result = np.geomspace(start=first_bin_mid, stop=last_bin_mid, num=self.n)
            except AttributeError:
                # Python3.2 require numpy older than 1.2, such versions of numpy doesn't have geomspace function
                # in such case we calculate geometrical binning manually
                result = np.exp(np.linspace(start=np.log(first_bin_mid),
                                            stop=np.log(last_bin_mid),
                                            num=self.n))
            return result
        else:
            return None


class AxisId(IntEnum):
    x = 0
    y = 1
    z = 2
    diff1 = 3
    diff2 = 4


@dataclass
class AxisDescription:
    """Single axis description"""

    name: str
    unit: str



from collections import namedtuple
import logging

import numpy as np

from pymchelper.axis import MeshAxis
from pymchelper.page import Page
from pymchelper.readers.shieldhit.reader_base import SHReader, mesh_unit_and_name, _bintyp, _get_detector_unit
from pymchelper.shieldhit.detector.detector_type import SHDetType
from pymchelper.shieldhit.detector.estimator_type import SHGeoType

logger = logging.getLogger(__name__)


class SHReaderBin2010(SHReader):
    """
    Binary format reader from 0.1 <= version <= 0.6
    """
    def read_header(self, estimator):
        logger.info("Reading header: " + self.filename)

        estimator.tripdose = 0.0
        estimator.tripntot = -1

        # effective read
        # first figure out if this is a VOXSCORE card
        header_dtype = np.dtype([('__fo1', '<i4'), ('geotyp', 'S10')])
        header = np.fromfile(self.filename, header_dtype, count=1)
        if not header:
            print("File {:s} has unknown format".format(self.filename))
            return None

        if 'VOXSCORE' in header['geotyp'][0].decode('ascii'):
            header_dtype = np.dtype([('__fo1', '<i4'),     # 0x00
                                     ('geotyp', 'S10'),    # 0x04
                                     ('__fo2', '<i4'),     # 0x0E
                                     ('__fo3', '<i4'),     # 0x12
                                     ('nstat', '<i4'),     # 0x16 : nstat
                                     ('__fo4', '<i4'),     # 0x1A
                                     ('__foo1', '<i4'),    # 0x1E
                                     ('tds', '<f4'),       # 0x22 : tripdose
                                     ('__foo2', '<i4'),    # 0x26
                                     ('__foo3', '<i4'),    # 0x2A
                                     ('tnt', '<i8'),       # 0x2E : tripntot
                                     ('__foo4', '<i4'),    # 0x36
                                     ('__fo5', '<i4'),     # 0x3A
                                     # DET has 8x float64
                                     ('det', ('<f8', 8)),  # 0x3E : DET
                                     ('__fo6', '<i4'),     # 0x7E
                                     ('__fo7', '<i4'),     # 0x82
                                     # IDET has 11x int32
                                     ('idet', '<i4', 11),  # 0x86 : IDET
                                     ('__fo8', '<i4'),     # 0xB2
                                     ('reclen', '<i4')])   # 0xB6
            # payload starts at 0xBA (186)
            estimator.payload_offset = 186
        else:
            # first figure out the length.
            header_dtype = np.dtype([('__fo1', '<i4'),
                                     ('geotyp', 'S10'),
                                     ('__fo2', '<i4'),
                                     ('__fo3', '<i4'),
                                     ('nstat', '<i4'),
                                     ('__fo4', '<i4'),
                                     ('__fo5', '<i4'),
                                     # DET has 8x float64
                                     ('det', ('<f8', 8)),  # DET
                                     ('__fo6', '<i4'),
                                     ('__fo7', '<i4'),
                                     # IDET has 11x int32
                                     ('idet', '<i4', 11),  # IDET
                                     ('__fo8', '<i4'),
                                     ('reclen', '<i4')])
            # payload starts at 0x9E (158)
            estimator.payload_offset = 158

        header = np.fromfile(self.filename, header_dtype, count=1)
        estimator.rec_size = header['reclen'][0] // 8

        if 'VOXSCORE' in header['geotyp'][0].decode('ascii'):
            estimator.tripdose = header['tds'][0]
            estimator.tripntot = header['tnt'][0]

        # map 10-elements table to namedtuple, for easier access
        # here is description of IDET table, assuming fortran-style numbering
        # (arrays starting from 1)
        # IDET(1) : Number of bins in first dimension. x or r or zones
        # IDET(2) : Number of bins in snd dimension, y or theta
        # IDET(3) : Number of bins in thrd dimension, z
        # IDET(4) : Particle type requested for scoring
        # IDET(5) : Detector type (see INITDET)
        # IDET(6) : Z of particle to be scored
        # IDET(7) : A of particle to be scored (only integers here)
        # IDET(8) : Detector material parameter
        # IDET(9) : Number of energy/amu (or LET) differential bins,
        #            negative if log.
        # IDET(10): Type of differential scoring, either LET, E/amu
        #            or polar angle
        # IDET(11): Starting zone of scoring for zone scoring
        DetectorAttributes = namedtuple('DetectorAttributes',
                                        ['dim_1_bins', 'dim_2_bins',
                                         'dim_3_bins',
                                         'particle_type', 'det_type',
                                         'particle_z', 'particle_a',
                                         'det_material',
                                         'diff_bins_no', 'diff_scoring_type',
                                         'starting_zone'])
        det_attribs = DetectorAttributes(*header['idet'][0])

        nx = det_attribs.dim_1_bins
        ny = det_attribs.dim_2_bins
        nz = det_attribs.dim_3_bins

        # DET(1-3): start positions for x y z or r theta z
        # DET(4-6): stop positions for x y z or r theta z
        # DET(7)  : start differential grid
        # DET(8)  : stop differential grid
        estimator.det = header['det']
        estimator.particle = det_attribs.particle_type

        try:
            estimator.geotyp = SHGeoType[header['geotyp'][0].decode('ascii').strip().lower()]
        except Exception:
            estimator.geotyp = SHGeoType.unknown
        estimator.number_of_primaries = header['nstat'][0]

        if estimator.geotyp not in {SHGeoType.zone, SHGeoType.dzone}:
            xmin = header['det'][0][0]
            ymin = header['det'][0][1]
            zmin = header['det'][0][2]

            xmax = header['det'][0][3]
            ymax = header['det'][0][4]
            zmax = header['det'][0][5]
        else:
            # special case for zone scoring, x min and max will be zone numbers
            xmin = det_attribs.starting_zone
            xmax = xmin + nx - 1
            ymin = 0.0
            ymax = 0.0
            zmin = 0.0
            zmax = 0.0

        if estimator.geotyp in {SHGeoType.plane, SHGeoType.dplane}:
            # special case for plane scoring, according to documentation we have:
            #  xmin, ymin, zmin = Sx, Sy, Sz (point on the plane)
            #  xmax, ymax, zmax = nx, ny, nz (normal vector)
            # to avoid situation where i.e. xmax < xmin (corresponds to nx < Sx)
            # we store only point on the plane
            estimator.sx, estimator.sy, estimator.sz = xmin, ymin, zmin
            estimator.nx, estimator.ny, estimator.nz = xmax, ymax, zmax
            xmax = xmin
            ymax = ymin
            zmax = zmin

        xunit, xname = mesh_unit_and_name(estimator, 0)
        yunit, yname = mesh_unit_and_name(estimator, 1)
        zunit, zname = mesh_unit_and_name(estimator, 2)

        estimator.x = MeshAxis(n=np.abs(nx), min_val=xmin, max_val=xmax, name=xname, unit=xunit, binning=_bintyp(nx))
        estimator.y = MeshAxis(n=np.abs(ny), min_val=ymin, max_val=ymax, name=yname, unit=yunit, binning=_bintyp(ny))
        estimator.z = MeshAxis(n=np.abs(nz), min_val=zmin, max_val=zmax, name=zname, unit=zunit, binning=_bintyp(nz))

        page = Page(estimator=estimator)
        page.dettyp = SHDetType(det_attribs.det_type)
        page.unit, page.name = _get_detector_unit(page.dettyp, estimator.geotyp)
        estimator.add_page(page)

        return True  # reading OK

    # TODO: we need an alternative list, in case things have been scaled with nscale, since then things
    # are not "/particle" anymore.
    def read_payload(self, estimator):
        logger.info("Reading data: " + self.filename)

        if estimator.geotyp == SHGeoType.unknown or estimator.pages[0].dettyp == SHDetType.none:
            logger.error("Unknown geotyp or dettyp")
            return None

        # next read the data:
        offset_str = "S" + str(estimator.payload_offset)
        record_dtype = np.dtype([('trash', offset_str),
                                 ('bin2', '<f8', estimator.rec_size)])
        record = np.fromfile(self.filename, record_dtype, count=-1)
        # BIN(*)  : a large array holding results. Accessed using pointers.
        estimator.pages[0].data_raw = np.array(record['bin2'][:][0])
        estimator.pages[0].error_raw = np.empty_like(estimator.pages[0].data_raw)

        logger.debug("Raw data: {}".format(estimator.pages[0].data_raw))

        estimator.file_counter = 1

        return True

    def read_data(self, estimator):
        if not self.read_header(estimator):
            logger.debug("Reading header failed")
            return None
        if not self.read_payload(estimator):
            logger.debug("Reading payload failed")
            return None
        estimator.file_format = 'bin2010'
        super(SHReaderBin2010, self).read_data(estimator)
        return True



import logging
from pathlib import Path
import struct

import numpy as np
import pymchelper

from pymchelper.page import Page
from pymchelper.shieldhit.detector.detector_type import SHDetType
from pymchelper.writers.writer import Writer

logger = logging.getLogger(__name__)


class MCPLWriter(Writer):
    """MCPL data writer"""

    def __init__(self, output_path: str, _):
        super().__init__(output_path)
        self.output_path = self.output_path.with_suffix(".mcpl")

    def write_single_page(self, page: Page, output_path: Path):
        """TODO"""
        logger.info("Writing page to: %s", str(output_path))

        # special case for MCPL data
        if page.dettyp == SHDetType.mcpl:

            # first part of the header
            header_bytes = "MCPL".encode('ascii')  # magic number
            header_bytes += "003".encode('ascii')  # version
            header_bytes += "L".encode('ascii')  # little endian
            header_bytes += struct.pack("<Q", page.data.shape[1])  # number of particles
            header_bytes += struct.pack("<I", 0)  # number of custom comments
            header_bytes += struct.pack("<I", 0)  # number of custom binary blobs
            header_bytes += struct.pack("<I", 0)  # user flags disabled
            header_bytes += struct.pack("<I", 0)  # polarisation disabled
            header_bytes += struct.pack("<I", 1)  # single precision for floats
            header_bytes += struct.pack("<i", 0)  # all particles have PDG code
            header_bytes += struct.pack("<I", 32)  # data length
            header_bytes += struct.pack("<I", 1)  # universal weight

            # second part of the header
            header_bytes += struct.pack("<d", 1)  # universal weight value

            # data arrays
            source_name = f"pymchelper {pymchelper.__version__}"
            header_bytes += struct.pack("<I", len(source_name))  # length of the source name
            header_bytes += source_name.encode('ascii')  # source name

            # particle data
            # need to fix the structure according to MCPL format
            # see https://mctools.github.io/mcpl/mcpl.pdf#nameddest=section.3
            # we use numpy to speed up the process for large arrays

            # Create a structured array with named fields, as some of the fields have different types
            dt = np.dtype([('x', np.float32), ('y', np.float32), ('z', np.float32), ('fp1', np.float32),
                           ('fp2', np.float32), ('uz', np.float32), ('time', np.float32), ('pdg', np.uint32)])
            data_bytes = np.empty(page.data.shape[1], dtype=dt)

            # Assign values to the fields
            data_bytes['x'] = page.data[1]
            data_bytes['y'] = page.data[2]
            data_bytes['z'] = page.data[3]
            data_bytes['fp1'] = page.data[4]  # ux by default
            data_bytes['fp2'] = page.data[5]  # uy by default
            data_bytes['time'] = 0
            data_bytes['pdg'] = page.data[0]

            ux2 = page.data[4]**2
            uy2 = page.data[5]**2
            uz2 = page.data[6]**2
            sign = np.ones_like(page.data[6], dtype=int)

            # find the maximum component of the direction vector
            # condition below defines the case where the maximum component is ux
            condition_1 = np.logical_and(ux2 >= uy2, ux2 > uz2)
            # condition below defines the case where the maximum component is uy
            condition_2 = np.logical_and(uy2 > ux2, uy2 > uz2)
            # by exclusion, the remaining case is where the maximum component is uz
            condition_3 = np.logical_not(np.logical_or(condition_1, condition_2))

            # fill the arrays according to the maximum component
            # lets start with the case where the maximum component is uz
            sign[(page.data[6] < 0) & condition_3] = -1

            # fill the arrays according to the maximum component ux
            sign[(page.data[4] < 0) & condition_1] = -1  # negative sign of ux
            data_bytes['fp1'][condition_1] = 1 / page.data[6][condition_1]  # 1/uz
            data_bytes['fp2'][condition_1] = page.data[5][condition_1]  # uy

            # fill the arrays according to the maximum component uy
            sign[(page.data[5] < 0) & condition_2] = -1  # sign of uy
            data_bytes['fp1'][condition_2] = page.data[4][condition_2]  # ux
            data_bytes['fp2'][condition_2] = 1 / page.data[6][condition_2]  # 1/uz

            data_bytes['uz'] = sign * page.data[7]

            data_bytes = data_bytes.tobytes()

            output_path.write_bytes(header_bytes + data_bytes)
            return



import logging
import os
from pathlib import Path
import sys
from typing import Optional

from pymchelper.simulator_type import SimulatorType


class MCEnvironment:
    """
    `MCEnvironment` subclasses are helpful to discover which MC engine (i.e. FLUKA or SHIELD-HIT12A) is being used
    they provide information about expected executable filename, by inspecting the path to executable filename
    (i.e. checking if it ends with `rfluka`) we can find corresponding code type
    """
    executable_filename = None


class FlukaEnvironment(MCEnvironment):
    """
    FLUKA Environment
    """
    executable_filename = 'rfluka'
    simulator_type = SimulatorType.fluka


class SH12AEnvironmentLinux(MCEnvironment):
    """
    SHIELD-HIT12A Environment for Linux
    """
    executable_filename = 'shieldhit'
    simulator_type = SimulatorType.shieldhit


class SH12AEnvironmentWindows(MCEnvironment):
    """
    SHIELD-HIT12A Environment for Windows
    """
    executable_filename = 'shieldhit.exe'
    simulator_type = SimulatorType.shieldhit


class TopasEnvironment(MCEnvironment):
    """TOPAS Environment"""

    executable_filename = 'topas'
    simulator_type = SimulatorType.topas


class SimulationSettings:
    """
    This class is responsible for keeping track of options for MC simulation:
      - location of the MC simulator executable
      - additional options provided by the user
      - location of the input files or directories
    Moreover this class performs automatic discovery of the MC input
    (i.e. whether this is SHIELD-HIT12A, Fluka or TOPAS input)
    """

    def __init__(self, input_path: str, simulator_type: Optional[SimulatorType]=None,
                 simulator_exec_path: str=None, cmdline_opts: str=None):
        # Input file or directory
        self.input_path = input_path

        # Set `self._mc_environment` to the proper `MCEnvironment` subclass
        # If `simulator_type` is provided by user, then we use it to set proper `MCEnvironment` subclass
        # Otherwise, we try to guess based on provided input path
        if simulator_type:
            if simulator_type == SimulatorType.shieldhit:
                if sys.platform == 'win32':
                    self._mc_environment = SH12AEnvironmentWindows
                else:
                    self._mc_environment = SH12AEnvironmentLinux
            elif simulator_type == SimulatorType.fluka:
                self._mc_environment = FlukaEnvironment
            elif simulator_type == SimulatorType.topas:
                self._mc_environment = TopasEnvironment
        else:
            self._mc_environment = self._discover_mc_engine(input_path)

        if not self._mc_environment:
            raise Exception(f"Unable to determine MC engine type for {input_path}")

        self.simulator_type = self._mc_environment.simulator_type

        # set `self.executable_path` to the value provided by user, or if it is missing
        # perform automatic discovery of the *location* of MC engine executable file by scanning PATH env. variable
        self.executable_path = simulator_exec_path
        if not self.executable_path:
            self.executable_path = self._discover_mc_exec_location(self._mc_environment.executable_filename)

        # set extra options (i.e. `--time 00:15:30 -v`) for the MC engine
        self.cmdline_opts = cmdline_opts if cmdline_opts else ''  # sanity check for None value
        # if extra options are provided, perform options validation
        # in case the options are not supported by MC engine, validation method will throw an exception
        if self.cmdline_opts:
            self._validate_cmdline_opt(self.cmdline_opts)

    def set_rng_seed(self, rng_seed):
        """
        This methods modifies command line options of the MC engine by setting (or overriding) the value of RNG seed
        TODO this method is specific to SH12A, more general should be added
        TODO add support for RNG seed provided as --seedofset, instead of -N
        """
        # transform option list from plain string to a list of values for easier manipulation
        options_list = self.cmdline_opts.split()

        # If RNG seed is missing on the option list, then the code below will set it to given value
        if '-N' not in options_list:
            self.cmdline_opts += " -N {:d}".format(rng_seed)
        # if RNG is present on the option list, then we override its value
        else:
            # in SHIELD-HIT12A RNG seed is specified by -N option
            index_of_rng_opt = options_list.index('-N')  # find index of '-N'
            options_list[index_of_rng_opt + 1] = str(rng_seed)  # override the value of current -N option
            self.cmdline_opts = ' '.join(options_list)  # reconstruct option string

    def set_no_of_primaries(self, number_of_primaries):
        """
        This methods modifies command line options of the MC engine by setting (or overriding) the number of primaries
        to be simulated by each of the parallel jobs
        TODO this method is specific to SH12A, more general should be added
        TODO add support for no of primaries provided as --nstat, instead of -n
        """
        # transform option list from plain string to a list of values for easier manipulation
        options_list = self.cmdline_opts.split()

        # If no of primaries is missing on the option list, then the code below will set it to given value
        if '-n' not in options_list:
            self.cmdline_opts += " -n {:d}".format(number_of_primaries)
        else:
            # see `set_rng_seed` for the logic
            index_of_prim_opt = options_list.index('-n')
            options_list[index_of_prim_opt + 1] = str(number_of_primaries)
            self.cmdline_opts = ' '.join(options_list)

    @staticmethod
    def _validate_cmdline_opt(cmdline_opts):
        """
        TODO this method is specific to SH12A, more general should be added
        """
        # transform option list from plain string to a list of values for easier manipulation
        options_list = cmdline_opts.split()

        # transform option list to a set to ease finding common part of unsupported and current options
        options_set = set(options_list)
        # set of options which cannot be overwritten by the user
        # these include locations of the input files which are fixed to the temporary workspace directories
        # generated by the `pymchelper` code
        unsupported = {'-b', '--beamfile', '-g', '--geofile', '-m', '--matfile', '-d', '--detectfile'}
        # raise an error if some of the unsupported option was provided by user (i.e. via -m option to `runmc` command)
        if options_set & unsupported:
            # TODO replace exception with warning and ignore such options  # skipcq: PYL-W0511
            raise SyntaxError("Unsupported option encountered: {:s}".format(",".join(options_set & unsupported)))

    @staticmethod
    def _discover_mc_engine(input_path_str: str):
        """
        Analyse the input path and based on its type set proper MC engine
        In case of failure return None
        """
        input_path = Path(input_path_str)
        # raise exception if invalid path is provided
        if not input_path.exists():
            raise Exception("Input path {:s} doesn't exists".format(input_path))

        # Fluka and TOPAS input files are provided as the single file
        # We return proper MC engine based on the file extension
        if input_path.is_file():
            if input_path_str.endswith('.inp'):
                return FlukaEnvironment
            if input_path_str.endswith('.txt'):
                return TopasEnvironment
        # SHIELD-HIT12A input is in the form of directory with multiple files
        elif input_path.is_dir():
            # We check if directory contains the required SHIELDHIT-12A input files
            if all(input_path.joinpath(file).exists() for file in ['beam.dat', 'mat.dat', 'geo.dat', 'detect.dat']):
                # in case pymchelper runs on Windows choose a SHIELDHIT-12A environment which is Windows specific
                # (executable file being `shieldhit.exe` instead of `shieldhit`)
                if sys.platform == 'win32':
                    return SH12AEnvironmentWindows
                return SH12AEnvironmentLinux
        return None

    @staticmethod
    def _discover_mc_exec_location(exec_filename):
        """
        Scans PATH variable for the possible location of the MC engine exec_filename.
        Works on POSIX (Linux and MacOSX) and Windows systems
        """

        # PATH variable contains list of directories being separated by : on POSIX systems
        # and by ; on Windows systems. Here we set proper separator
        separator_char = ':'
        if sys.platform == 'win32':
            separator_char = ';'

        list_of_directories = []
        # loop over all directories in PATH env variable
        for path_directory_entry in os.environ['PATH'].split(separator_char):

            logging.debug("Inspecting {:s}".format(path_directory_entry))

            # check if PATH entry is a proper directory
            # some malicious users can add a files to PATH as well (although it doesn't make any sense)
            if os.path.isdir(path_directory_entry):

                # scan a PATH entry and loop over all filenames inside PATH entry directory using `os.listdir`
                if exec_filename in os.listdir(path_directory_entry):

                    # add a directories for which we encounter `exec_filename`
                    # (MC engine filename, like `rfluka` or `shieldhit`)
                    list_of_directories.append(path_directory_entry)

        # raise exception if MC engine not found
        if not list_of_directories:
            raise Exception("Executable {:s} not found in PATH ({:s})".format(exec_filename, ",".join(sys.path)))

        # if `exec_filename` is found in multiple location, return full path to the first location (directory+filename)
        return os.path.join(list_of_directories[0], exec_filename)

    def __str__(self):
        """
        Dump all settings into a string
        """
        result = "{executable_path:s} {cmdline_opts:s}".format(
            executable_path=os.path.abspath(self.executable_path),
            cmdline_opts=self.cmdline_opts if self.cmdline_opts else ''
        )
        return result


class SH12ASettings(SimulationSettings):
    """
    TODO
    """



#!/usr/bin/env python

import argparse
import glob
import logging
import sys

from pymchelper.estimator import ErrorEstimate
from pymchelper.input_output import convertfromlist, convertfrompattern
from pymchelper.writers.common import Converters
from pymchelper.writers.plots import ImageWriter, PlotAxis

logger = logging.getLogger(__name__)


def add_default_options(parser):
    import pymchelper
    parser.add_argument('input', help='input filename, file list or pattern', type=str)
    parser.add_argument('output', help='output filename or directory', nargs='?')
    parser.add_argument('--many', help='automatically merge data from various sources', action="store_true")
    parser.add_argument('-a', '--nan', help='ignore NaN in averaging', action="store_true")
    parser.add_argument('-e', '--error',
                        help='type of error estimate to add (default: ' + ErrorEstimate.stderr.name + ')',
                        default=ErrorEstimate.stderr.name,
                        choices=[x.name for x in ErrorEstimate],
                        type=str)
    parser.add_argument('-n', '--nscale', help='scale with number of primaries N.', default=1, type=float)
    parser.add_argument('-v',
                        '--verbose',
                        action='count',
                        default=0,
                        help='give more output. Option is additive, and can be used up to 3 times')
    parser.add_argument('-q', '--quiet', action='count', default=0, help='be silent')
    parser.add_argument('-V', '--version', action='version', version=pymchelper.__version__)


def main(args=None):
    if args is None:
        args = sys.argv[1:]
    import os

    import pymchelper

    _progname = os.path.basename(sys.argv[0])
    _helptxt = 'Universal converter for FLUKA and SHIELD-HIT12A output files.'
    _epitxt = f"Type '{_progname} <converter> --help' for help on a specific converter."

    parser = argparse.ArgumentParser(description=_helptxt, epilog=_epitxt)

    subparsers = parser.add_subparsers(dest='command', metavar='converter')

    parser_txt = subparsers.add_parser(Converters.txt.name, help='converts to plain txt file')
    add_default_options(parser_txt)

    parser_image = subparsers.add_parser(Converters.image.name, help='converts to PNG images')
    add_default_options(parser_image)
    axis_names = [x.name for x in PlotAxis]
    parser_image.add_argument('-l', '--log',
                              help='set logscale for plot axis',
                              nargs='+',
                              choices=axis_names,
                              default={},
                              type=str)
    parser_image.add_argument("--colormap",
                              help='image color map, see https://matplotlib.org/stable/tutorials/colors/colormaps.html '
                                   'for list of possible options (default: ' + ImageWriter.default_colormap + ')',
                              default=ImageWriter.default_colormap,
                              type=str)

    parser_excel = subparsers.add_parser(Converters.excel.name, help='converts to MS Excel file')
    add_default_options(parser_excel)

    parser_hdf = subparsers.add_parser(Converters.hdf.name, help='converts to HDF file')
    add_default_options(parser_hdf)

    parser_json = subparsers.add_parser(Converters.mcpl.name, help='converts to MCPL phase space file')
    add_default_options(parser_json)

    parser_json = subparsers.add_parser(Converters.json.name, help='converts to JSON file')
    add_default_options(parser_json)

    parser_plotdata = subparsers.add_parser(Converters.plotdata.name, help='converts to plot data')
    add_default_options(parser_plotdata)

    parser_inspect = subparsers.add_parser(Converters.inspect.name, help='prints metadata')
    add_default_options(parser_inspect)
    parser_inspect.add_argument('-d', '--details',
                                help='print detailed information about data attribute',
                                action="store_true")

    parser_sparse = subparsers.add_parser(Converters.sparse.name, help='converts to sparse matrix format')
    add_default_options(parser_sparse)
    parser_sparse.add_argument("--threshold",
                               help='only values greater than threshold are saved',
                               type=float,
                               default=0.0)

    parser_tripcube = subparsers.add_parser(Converters.tripcube.name, help='converts to TRiP98 data cube')
    add_default_options(parser_tripcube)

    parser_tripddd = subparsers.add_parser(Converters.tripddd.name, help='converts to TRiP98 DDD file')
    add_default_options(parser_tripddd)
    parser_tripddd.add_argument("--energy",
                                help='energy of the beam [MeV/amu] (guess from data if option missing)',
                                type=float)
    parser_tripddd.add_argument("--projectile",
                                help='projectile (guess from data if option missing)',
                                type=str)
    parser_tripddd.add_argument("--ngauss",
                                help='number of Gaussian functions to fit (default: 0)',
                                choices=(0, 1, 2),
                                default=0,
                                type=int)

    parser.add_argument('-V', '--version', action='version', version=pymchelper.__version__)

    parsed_args = parser.parse_args(args)

    verbose_flag = getattr(parsed_args, 'verbose', 0)
    if verbose_flag == 1:
        logging.basicConfig(level=logging.INFO)
    elif verbose_flag > 1:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig()

    status = 0
    if parsed_args.command is not None:
        # TODO add filename discovery
        files = sorted(glob.glob(parsed_args.input))
        if not files:
            logger.error('File %s does not exist: ', parsed_args.input)

        # check if output should be interpreted as a filename
        if not parsed_args.many and len(files) == 1:
            output_file = parsed_args.output
        else:
            output_file = None

        if parsed_args.output is not None and output_file is None:
            output_dir = parsed_args.output
            # check if output directory exists
            if output_dir and not os.path.exists(output_dir):
                logger.warning("Directory %s does not exist, creating.", output_dir)
                os.makedirs(output_dir)
        else:
            output_dir = '.'

        parsed_args.error = ErrorEstimate[parsed_args.error]

        if parsed_args.many:
            status = convertfrompattern(parsed_args.input, output_dir,
                                        converter_name=parsed_args.command, options=parsed_args,
                                        error=parsed_args.error, nan=parsed_args.nan)
        else:
            status = convertfromlist(parsed_args.input,
                                     error=parsed_args.error, nan=parsed_args.nan, outputdir=output_dir,
                                     converter_name=parsed_args.command, options=parsed_args, outputfile=output_file)

    return status


if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))
